# -*- coding: UTF-8 -*-
from .lexer import WhereStatementLexer
from .table import BASE, FIELDS, FIELDS_MAP


class CredsSQLiteDB:
    def __init__(self, path, logger=None, **kwargs):
        from tinyscript import logging
        self.path = path
        self.__logger = logger or logging.nullLogger
    
    def __enter__(self):
        self.connect()
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
    
    def _exec(self, cursor, statement, *args, **kwargs):
        self.__logger.debug(f"{statement} - {args[0]}" if len(args) > 0 and isinstance(args[0], tuple) else statement)
        cursor.execute(statement, *args, **kwargs)
    
    def close(self):
        try:
            self.__conn.close()
        except AttributeError:
            pass
    
    def connect(self):
        from sqlite3 import connect
        self.__conn = connect(str(self.path))
    
    def count(self, *fields):
        cursor = self.__conn.cursor()
        self._exec(cursor, f"SELECT {', '.join(fields)}, COUNT(*) AS count FROM {BASE} GROUP BY {', '.join(fields)}")
        for row in cursor.fetchall():
            yield row
    
    def create(self):
        fields = ", ".join(f"{f} {d['type']}" for f, d in FIELDS.items())
        primary_keys = ", ".join(f for f, d in FIELDS.items() if d.get('primary', False))
        cursor = self.__conn.cursor()
        self._exec(cursor, f"DROP TABLE IF EXISTS {BASE}")
        self._exec(cursor, f"CREATE TABLE {BASE} ({fields}, PRIMARY KEY ({primary_keys}))")
        self.__conn.commit()
    
    def distinct(self, *fields):
        multi = len(fields) > 1
        fields = ", ".join(fields)
        cursor = self.__conn.cursor()
        self._exec(cursor, f"SELECT COUNT(*) AS count FROM (SELECT {fields} FROM {BASE} GROUP BY {fields})" if multi \
                      else f"SELECT COUNT(DISTINCT {fields}) AS {fields}_count FROM {BASE}")
        for row in cursor.fetchall():
            yield row
    
    def ingest(self, **record):
        data, primary_keys = [{}], [f for f, d in FIELDS.items() if d.get('primary', False)]
        # create data from the input record, handling data expansion where relevant
        for k, v in record.items():
            k = FIELDS_MAP.get(k, k)
            f = FIELDS[k].get('format', lambda x: x)
            v = FIELDS[k].get('expand', lambda x: x)(v)
            if isinstance(v, list):
                if k in primary_keys:
                    from itertools import product
                    old, data = tuple(data), []
                    for d, v2 in product(old, v):
                        d2 = {k: v for k, v in d.items()}
                        d2[k] = f(v2)
                        data.append(d2)
                else:
                    v2 = ",".join(map(f, v))
                    for d in data:
                        d[k] = v2
            else:
                for d in data:
                    d[k] = str(f(v))
        # populate values for primary keys
        for f in primary_keys:
            for d in data:
                if f in d:
                    continue
                d[f] = ""
        # now save to the database
        for d in data:
            fields, marks = ", ".join(d.keys()), ", ".join(len(d) * ['?'])
            update = ", ".join(f"{f} = excluded.{f}" for f in d.keys() if f not in primary_keys)
            update = "DO NOTHING" if update == "" else f"DO UPDATE SET {update}"
            self.__conn.cursor().execute(f"INSERT INTO {BASE} ({fields}) VALUES ({marks}) ON CONFLICT" \
                                         f"({', '.join(primary_keys)}) {update}", tuple(d.values()))
        self.__conn.commit()
    
    def select(self, where_statement, fields=None):
        where, values = WhereStatementLexer.prepare(where_statement)
        cursor = self.__conn.cursor()
        self._exec(cursor, f"SELECT {'*' if fields is None else ', '.join(fields)} FROM {BASE} WHERE {where}", values)
        for row in cursor.fetchall():
            yield row




# -*- coding: UTF-8 -*-
from argparse import Action, SUPPRESS as _SUP
from tinyscript import *

from .__info__ import __author__, __copyright__, __email__, __license__, __version__
from .db import CredentialsDB


__source__    = "https://github.com/dhondta/searchpass"
__doc__       = """
This tool aims to search for default passwords of common devices based on criteria like the vendor or the model.
It works by caching the whole lists of known default passwords downloaded from various sources (relying on pybots ;
 including CIRTnet, DataRecovery, PasswordDB, RouterPasswd or even SaynamWeb) to perform searches locally.
"""
__examples__  = [
    "--update",
    "--passwords",
    "--stats",
    "--query \"username='user'",
    "--query \"username LIKE \\\"Admin%%\\\"\" --passwords",
]

DB_ACTIONS = {
    'reset':  "remove cached credentials databases",
    'show':   "show records of credentials databases",
    'stats':  "get statistics on credentials databases",
    'update': "update credentials databases",
}


def custom_action(method):
    class _Action(Action):
        """ Custom action for applying a method of searchpass.db.CredentialsDB. """
        def __init__(self, option_strings, dest=_SUP, help=None):
            super(_Action, self).__init__(option_strings=option_strings, dest=_SUP, default=_SUP, nargs=0, help=help)
        
        def __call__(self, parser, namespace, values, option_string=None):
            namespace._action = method
    return _Action


def main():
    """ Tool's main function """
    from shutil import which
    if not which("visidata"):
        del DB_ACTIONS['show']
    for a in DB_ACTIONS.keys():
        cls = globals()[f'_{a.capitalize()}Action'] = custom_action(a)
        parser.register('action', a, cls)
    so = parser.add_argument_group("search options")
    fmt = so.add_mutually_exclusive_group()
    fmt.add_argument("-e", "--empty", action="store_true", help="include empty username or password")
    fmt.add_argument("--passwords", action="store_true", help="get passwords only")
    so.add_argument("-q", "--query", help="search query")
    fmt.add_argument("--usernames", action="store_true", help="get usernames only")
    dba = parser.add_argument_group("action arguments")
    for a, h in DB_ACTIONS.items():
        dba.add_argument(f"--{a}", action=a, help=h)
    initialize()
    args.logger = logger
    credsdb = CredentialsDB(**vars(args))
    m = getattr(args, "_action", None)
    if m is not None:
        getattr(credsdb, m)(**vars(args))
    else:
        creds = credsdb.search(**vars(args))
        if len(creds) == 0:
            logger.warning("No credential found")
        elif args.usernames:
            for u in sorted(set(creds.keys())):
                print(u)
        elif args.passwords:
            pswds = []
            for plst in creds.values():
                pswds.extend(plst)
            for p in sorted(set(pswds)):
                print(p)
        else:
            for user, pswds in sorted(creds.items(), key=lambda x: x[0].lower()):
                for pswd in set(pswds):
                    print("%s:%s" % (user, pswd))
    return 0




# -*- coding: UTF-8 -*-
from pybots.apis.security.default_credentials import *
from pybots.apis.security.default_credentials import __all__ as _apis
from pybots.core.utils.api import APIError
from tinyscript import *
from tinyscript.report import *

from .sql import CredsSQLiteDB


__all__ = ["CredentialsDB"]

_DEPRECATED = ["SaynamWebAPI"]
APIS = [a for a in _apis if a not in _DEPRECATED]
DB = ts.Path("~/.searchpass.db", expand=True)


def _render(*elements):
    from rich.box import SIMPLE_HEAD
    from rich.console import Console
    from rich.markdown import Heading, Markdown
    from rich.style import Style
    from rich.table import Table as RichTable
    from rich.text import Text as RichText
    code.replace(Heading.__rich_console__, "text.justify = \"center\"", "")
    for e in elements:
        if hasattr(e, "md"):
            if isinstance(e, Table):
                opt = {'show_header': True, 'header_style': "bold", 'highlight': True}
                if e.title is not None:
                    opt['title'] = e.title
                    opt['title_justify'] = "left"
                    opt['title_style'] = Style(bold=True, color="bright_yellow", italic=False)
                if getattr(e, "borderless", True):
                    opt['box'] = SIMPLE_HEAD
                table = RichTable(**opt)
                for col in e.column_headers:
                    table.add_column(col, justify="center")
                for row in e.data:
                    if not all(set(str(cell).strip()) == {"-"} for cell in row):
                        table.add_row(*[RichText(str(cell), justify="left") for cell in row])
                Console(markup=False).print(table)
            elif isinstance(e, Section):
                Console().print(Markdown(e.md()), style=Style(bold=True, color="bright_cyan", italic=False))
            else:
                Console().print(Markdown(e.md()))
        elif e is not None:
            Console().print(Markdown(e))


class CredentialsDB:
    path = DB if DB.exists() else ts.Path(__file__).dirname.joinpath("searchpass.db")
    
    def __init__(self, logger=None, verbose=False, **kw):
        """ Loads and aggregates the records for every source of default credentials from the JSON files. """
        self.__logger, self.__verbose = logger or logging.nullLogger, verbose
    
    def reset(self, **kw):
        """ Resets cached JSON files. """
        DB.remove(False)
    
    def search(self, query=None, empty=False, **kw):
        """ Searches based on the given query in the given password dictionaries. """
        creds = {}
        with CredsSQLiteDB(self.path, logger=self.__logger) as db:
            try:
                for row in db.select(query, ("username", "password")):
                    user, pswd = row[0], row[1]
                    if user is None or pswd is None:
                        continue
                    if not empty and (user == "" or pswd == ""):
                        continue
                    m = re.match(r"^<<(.*?)>>$", pswd)
                    if m:
                        self.__logger.warning(f"Dynamic password found for username '{user}': {m.group(1)}")
                    creds.setdefault(user, [])
                    if pswd not in creds[user]:
                        creds[user].append(pswd)
            except SyntaxError as e:
                self.__logger.error(f"Bad query ({e})")
        return creds
    
    def show(self, **kw):
        """ Shows available credentials per source. """
        ts.execute(["visidata", str(self.path)])
    
    def stats(self, **kw):
        """ Computes statistics of the currently available passwords databases. """
        with CredsSQLiteDB(self.path, logger=self.__logger) as db:
            counts = [row for row in db.count("source")]
        _render(Section("Counts by source"), Table(counts, column_headers=["Source", "Count"]))
        stats = []
        with CredsSQLiteDB(self.path, logger=self.__logger) as db:
            stats.append([f"#unique credentials", next(db.distinct("username", "password"))[0]])
            for field in ["password", "username", "vendor"]:
                stats.append([f"#unique {field}s", next(db.distinct(field))[0]])
        _render(Section("Statistics"), Table(stats, column_headers=["Name", "Metric"]))
    
    def update(self, **kw):
        """ Downloads the records for each source of default passwords separately and caches them to JSON files. """
        self.__logger.debug("Recreating DB at {self.path}...")
        with CredsSQLiteDB(DB, logger=self.__logger) as db:
            db.create()
        for api_name in APIS:
            name = api_name[:-3]
            self.__logger.info(f"Downloading default credentials from {name}...")
            with globals()[api_name](verbose=self.__verbose) as api, CredsSQLiteDB(DB, logger=self.__logger) as db:
                for vendor in api.vendors:
                    try:
                        for record in api.credentials(vendor)['data']:
                            db.ingest(source=name, vendor=vendor, **record)
                    except APIError as e:
                        self.__logger.warning(e)
                self.__logger.debug("Saving to DB...")
        CredentialsDB.path = DB




# -*- coding: UTF-8 -*-
"""Searchpass package information.

"""
import os
from datetime import datetime


__author__    = "Alexandre D'Hondt"
__copyright__ = f"© 2021-{datetime.now().year} A. D'Hondt"
__email__     = "alexandre.dhondt@gmail.com"
__license__   = "GPLv3 (https://www.gnu.org/licenses/gpl-3.0.fr.html)"
__source__    = "https://github.com/dhondta/searchpass"

with open(os.path.join(os.path.dirname(__file__), "VERSION.txt")) as f:
    __version__ = f.read().strip()




# -*- coding: UTF-8 -*-
from .table import FIELDS


TOKEN_SPEC = [
    ("FIELD",    r"\b(?:[a-z](_?[a-z])*)\b"),                      # Allowed field names
    ("LIKE", r"\b(?:NOT|LIKE)\b"),                                 # LIKE operator (with NOT)
    ("OPERATOR", r"=|!=|<>"),                                      # Allowed operators
    ("STRING",   r"(?P<quote>['\"])(?:(?!\1).|\\.)*?(?P=quote)"),  # String in single quotes
    ("AND",      r"\bAND\b"),                                      # AND operator
    ("OR",       r"\bOR\b"),                                       # OR operator
    ("LPAREN",   r"\("),                                           # Left parenthesis
    ("RPAREN",   r"\)"),                                           # Right parenthesis
    ("SKIP",     r"[ \t]+"),                                       # Skip whitespace
]


class WhereStatementLexer:
    def __init__(self, text):
        self.__text = text
        self.__tokens = WhereStatementLexer.tokenize(text)
        self.__valid = WhereStatementLexer.validate(text)
    
    @property
    def text(self):
        return self.__text
    
    @property
    def tokens(self):
        return self.__tokens
    
    @property
    def valid(self):
        return self.__valid
    
    @staticmethod
    def prepare(text):
        tokens = WhereStatementLexer.tokenize(text)
        WhereStatementLexer.parse(tokens)
        return " ".join("?" if kind == "STRING" else value for kind, value in tokens), \
               tuple(value.strip("'\"") for kind, value in tokens if kind == "STRING")
    
    @staticmethod
    def parse(text_or_tokens):
        tokens = WhereStatementLexer.tokenize(text_or_tokens) if isinstance(text_or_tokens, str) else text_or_tokens
        stack, i = [], 0
        while i < len(tokens):
            kind, value = tokens[i]
            if kind == "FIELD":
                if value not in FIELDS.keys():
                    raise SyntaxError(f"Unknown field: {value}")
                # expect FIELD OP STRING
                if i + 2 < len(tokens) and (tokens[i+1][0] == "OPERATOR" or tokens[i+1][1] == "LIKE") and \
                   tokens[i+2][0] == "STRING":
                    stack.append((kind, value))
                    stack.append(tokens[i+1])
                    stack.append(tokens[i+2])
                    i += 3
                elif i + 3 < len(tokens) and tokens[i+1][0] == "LIKE" and tokens[i+1][1] == "NOT" and \
                   tokens[i+2][0] == "LIKE" and tokens[i+3][0] == "STRING":
                    stack.append((kind, value))
                    stack.append(tokens[i+1])
                    stack.append(tokens[i+2])
                    stack.append(tokens[i+3])
                    i += 4
                else:
                    raise SyntaxError("Expected (OPERATOR | NOT LIKE) STRING after FIELD")
            elif kind in {"AND", "OR"}:
                # process AND with higher precedence than OR
                while stack and stack[-1][0] == "AND" and kind == "OR":
                    op = stack.pop()
                    rhs = stack.pop()
                    lhs = stack.pop()
                    stack.append(("EXPR", (lhs, op, rhs)))
                stack.append((kind, value))
                i += 1
            elif kind in {"LPAREN", "RPAREN"}:
                stack.append((kind, value))
                i += 1
            else:
                raise SyntaxError(f"Unexpected token: {value}")
        # handle remaining operators in the stack
        while len(stack) > 1:
            rhs = stack.pop()
            op = stack.pop()
            lhs = stack.pop()
            stack.append(("EXPR", (lhs, op, rhs)))
        return stack[0] if stack else None
    
    @staticmethod
    def tokenize(text):
        import re
        regex, tokens = re.compile("|".join(f"(?P<{name}>{pattern})" for name, pattern in TOKEN_SPEC)), []
        for match in regex.finditer(text):
            kind = match.lastgroup
            value = match.group()
            if kind == "SKIP":
                continue
            tokens.append((kind, value))
        return tokens
    
    @staticmethod
    def validate(text_or_tokens):
        try:
            expr = WhereStatementLexer.parse(text_or_tokens)
            if expr is not None:
                raise SyntaxError(f"Unparsed expression: {expr[1]}")
            return True
        except:
            return False




# -*- coding: UTF-8 -*-
BASE = "credentials"
FIELDS = {
    'source':       {'type': "TEXT NOT NULL", 'primary': True},
    'vendor':       {'type': "TEXT NOT NULL", 'primary': True},
    'model':        {'type': "TEXT NOT NULL", 'primary': True},
    'version':      {'type': "TEXT", 'primary': True},
    'access_type':  {'type': "TEXT", 'format': lambda x: x.lower().replace("/", ",")},
    'access_level': {'type': "TEXT"},
    'notes':        {'type': "TEXT"},
    'username':     {'type': "TEXT", 'primary': True},
    'password':     {'type': "TEXT", 'primary': True, 'format': lambda p: _pswd_formatter(p),
                                                      'expand': lambda p: _pswd_splitter(p)},
}
FIELDS_MAP = {
    'manufacturer': "vendor",
    'name':         "model",
    'revision':     "version",
    'access-type':  "access_type",
    'method':       "access_type",
    'protocol':     "access_type",
    'access-level': "access_level",
    'level':        "access_level",
    'privileges':   "access_level",
    'doc':          "notes",
    'user':         "username",
    'user-id':      "username",
    'pass':         "password",
    'passwd':       "password",
    'pswd':         "password",
}


def _pswd_formatter(password):
    if any(pattern in password for pattern in \
       ["just hit ENTER", "leave blank", "no password", "(blank)", "(no pw)", "(none by default)"]):
        return ""
    if any(pattern in password.lower() for pattern in \
       ["ip address", "serial number", "model#", "mac address", "unknown", "digits of "]):
        from re import sub
        password = sub(r"^The\s", "", password.strip(" ."))
        return f"<<{password}>>"
    return password


def _pswd_splitter(password, separators=(" / ", " or ", "/")):
    from itertools import chain
    from re import sub
    password = sub(r"\s\(or\sblank\)", " (blank)", password)
    for sep in separators:
        if sep in password:
            tokens = password.split(sep)
            return list(chain.from_iterable((x if isinstance(x, list) else [x] for x in \
                        [_pswd_splitter(p, ("/", )) for p in password.split(sep)])))
    return password

